#include <pololu/orangutan.h>

#define SVP_GET_ID 0
#define SVP_SET_SERVO 1
#define SVP_SET_DIGITAL_STREAM 2
#define SVP_SET_DIGITAL_HIGH 3
#define SVP_SET_DIGITAL_LOW 4
#define SVP_SET_ANALOG_STREAM 5
#define SVP_DIGITAL_STREAM 6
#define SVP_ANALOG_STREAM 7
#define SVP_SET_MOTOR 8
#define SVP_SET_BUTTON 9
#define SVP_SET_BATTERY_STREAM 10
#define SVP_BATTERY_STREAM 11
#define SVP_SET_LCD 12
#define SVP_PLAY 13


#define PIN_IS_OUTPUT 0
#define PIN_IS_INPUT 1

// holds what packet type is being sent by the PC
unsigned int command;

// holds what pin/channel the command is referring to (not always used)
unsigned int channel;

//holds the value of the command
unsigned int value;

// used to determine when the digital values change so we don't send duplicate information
unsigned long lastDigital=0;
// used to determine when analog values change
unsigned int lastAnalog[13];
// holds if a pin is in output or input mode
unsigned char pinModes[4];

//motor1 speed
unsigned int M1Speed = 0;
//motor2 speed
unsigned int M2Speed = 0;

// holds the current button state, again to know when to transmit new information
unsigned char lastButtonState[3] = {0, 0, 0};
// current battery value
unsigned int battery=0;
// last battery value
unsigned int lastBattery=0;
// indicates that battery values are to be sent to the PC
unsigned int batteryStream=0;
// holds which analog channels are to be sent to the PC
unsigned int streamAnalog=0;
// holds which digital channels are to be sent to the PC
unsigned long streamDigital=0;

// serial receive buffer: A ring buffer that we will use to receive bytes on USB_COMM.
char receive_buffer[32];
// receive_buffer_position: This variable will keep track of which bytes in the receive buffer
// we have already processed.
unsigned char receive_buffer_position = 0;

// serial send buffer.
unsigned char send_buffer[32];

// holds the notes.
unsigned char notes[1024];

void initialize()
{
	int i;
	streamDigital=0;
	streamAnalog=0;
	lastDigital=-1;
	for (i=0;i<12;i++) lastAnalog[i]=-1;
	for (i=0;i<4;i++) pinModes[i]=PIN_IS_OUTPUT;

	lastBattery=0;

	lastButtonState[0]=0;
	lastButtonState[1]=0;
	lastButtonState[2]=0;
}

// wait_for_sending_to_finish:  Waits for the bytes in the send buffer to
// finish transmitting on USB_COMM.  We must call this before modifying
// send_buffer or trying to send more bytes, because otherwise we could
// corrupt an existing transmission.
void wait_for_sending_to_finish()
{
	while(!serial_send_buffer_empty(USB_COMM))
		serial_check();		// USB_COMM port is always in SERIAL_CHECK mode
}

// waits for and then grabs the current serial byte received
unsigned char getSerialByte()
{
	while(serial_get_received_bytes(USB_COMM) == receive_buffer_position)
		serial_check();

	unsigned char res = receive_buffer[receive_buffer_position];

	// Increment receive_buffer_position, but wrap around when it gets to
	// the end of the buffer.
	if (receive_buffer_position == sizeof(receive_buffer)-1)
		receive_buffer_position = 0;
	else
		receive_buffer_position++;

	return res;
}

// return a packet of information back to the PC
void writePacket()
{
	wait_for_sending_to_finish();

	send_buffer[0]=command|128;
	send_buffer[1]=channel;
	serial_send(USB_COMM, (char *)send_buffer, 2);
}

// return a packet of information with a value back to the PC
void writeValuePacket(int value)
{
	wait_for_sending_to_finish();

	send_buffer[0]=command|128;
	send_buffer[1]=channel;
	send_buffer[2]=value&127;
	send_buffer[3]=(value>>7)&127;
	send_buffer[4]=(send_buffer[0]^send_buffer[1]^send_buffer[2]^send_buffer[3])&127;

	serial_send(USB_COMM, (char *)send_buffer, 5);
}

void readPacket()
{
	// get header byte
	// 128 (bit 8) flag indicates a new command packet .. that
	// means the value bytes can never have 128 set!
	// next 7 bits are the command 0-7
	// next byte is the channel 0-32
	do
	{
		command = getSerialByte();
	}
	while ((command&128)==0);
	command^=128;

	channel = getSerialByte();
}

// reads a value from the PC typicall following a readPacket for some commands
int readValuePacket()
{
	unsigned int valueLow;
	unsigned int valueHigh;

	// wait for value low byte
	valueLow = getSerialByte();
	if (valueLow&128) return 0;

	// wait for value high byte
	valueHigh = getSerialByte();
	if (valueHigh&128) return 0;

	// wait for crc byte
	unsigned int crc = getSerialByte();
	if (crc&128) return 0;

	if (crc!=(((128|command)^channel^valueLow^valueHigh)&127)) return 0;

	value = valueLow|(valueHigh<<7);

	//print(" ");
	//print_long(value);

	return 1;
}

// Reads a string of text from the PC and formats for display on the LCD
int readLCDPacket()
{
	unsigned char text[64];

	unsigned int check=(128|command)^channel;
	int i=-1;
	do
	{
		i++;
		text[i] = getSerialByte();
		check ^= text[i];
	}
	while (text[i]);

	// wait for crc byte
	unsigned int crc = getSerialByte();
	if (crc&128) return 0;

	if (crc!=(check&127)) return 0;

	int row=0;
	lcd_goto_xy(0,0);
	for (i=0;text[i];i++)
	{
		if ((text[i]=='\n')||(text[i]=='\r'))
		{
			int p=i;
			while (p<16)
			{
				print(" ");
				p++;
			}
			lcd_goto_xy(0,1);
			row=1;
		}
		else
		if (text[i]=='\t')
			print("    ");
		else
		{
			if (i==16)
			{
				lcd_goto_xy(0,1);
				row=1;
			}
			print_character(text[i]);
		}
	}
	if (row)
	{
		while (i<32)
		{
			print(" ");
			i++;
		}
	}
	else
	{
		while (i<16)
		{
			print(" ");
			i++;
		}
		lcd_goto_xy(0,1);
		print("                ");
	}

	return 1;
}

// reads a string of notes from the PC and plays them. The notes are already in Orangutan SVP format
int readNotesPacket()
{
	unsigned int check=(128|command)^channel;
	int i=-1;
	do
	{
		i++;
		notes[i] = getSerialByte();
		check ^= notes[i];
	}
	while ((i<1024)&&(notes[i]));

	// wait for crc byte
	unsigned int crc = getSerialByte();
	if (crc&128) return 0;

	if (crc!=(check&127)) return 0;

    play(notes);

	return 1;
}

int main()
{
	// init servos ... do not forget to
	// - Connect PB3 to SA.
	// - Connect PB4 to SB.
 	// - Connect PC0 to SC.
 	// for the servos to work
 	const unsigned char demuxPins[] = {IO_B3, IO_B4, IO_C0}; // eight servos, B3=SA, B4=SB, C0=B5.
	servos_start(demuxPins, sizeof(demuxPins));

	set_servo_speed(0, 0);
	set_servo_speed(1, 0);
	set_servo_speed(2, 0);
	set_servo_speed(3, 0);
	set_servo_speed(4, 0);
	set_servo_speed(5, 0);
	set_servo_speed(6, 0);
	set_servo_speed(7, 0);

	serial_set_baud_rate(USB_COMM, 115200);

	// Start receiving bytes in the ring buffer.
	serial_receive_ring(USB_COMM, receive_buffer, sizeof(receive_buffer));

	initialize();
	clear();	// clear the LCD
	print("RR Gateway");

	while(1)
	{
		if (serial_get_received_bytes(USB_COMM) == receive_buffer_position)
			serial_check();
		else
		{
			readPacket();

			//lcd_goto_xy(0,1);
			//print_long(command);
			//print(" ");
			//print_long(channel);


			switch (command)
			{
				// init
			case  SVP_GET_ID:
				initialize();
				serial_send(USB_COMM, "OSVP", 4);
				break;
				// servo
			case  SVP_SET_SERVO:
				if ((channel>=0)&&(channel<8))
				{
					if (readValuePacket())
					{
						set_servo_target(channel, value);
						writeValuePacket(value);
					}
				}
				break;
				// motor
			case  SVP_SET_MOTOR:
				if ((channel>=0)&&(channel<2))
				{
					if (readValuePacket())
					{
						// switch to negative values
						if (value>255) value=value-16384;

						if (channel==0) M1Speed=value;
						if (channel==1) M2Speed=value;

						set_motors(M1Speed, M2Speed);
						writeValuePacket(value);
					}
				}
				break;
				//digital stream
			case  SVP_SET_DIGITAL_STREAM:
				if (readValuePacket())
				{
					streamDigital = value;
					writeValuePacket(value);
					lastDigital=-1;
				}
				break;
				//set digital high
			case  SVP_SET_DIGITAL_HIGH:
				if ((channel>=0)&&(channel<5))
				{
					if (pinModes[channel]!=PIN_IS_OUTPUT)
					{
						pinModes[channel]=PIN_IS_OUTPUT;
						if (streamDigital&((unsigned long)1<<channel))
							streamDigital^=(unsigned long)1<<channel;
					}
					set_digital_output(IO_D0+channel, HIGH);
					writePacket();
				}
				break;
				//set digital low
			case  SVP_SET_DIGITAL_LOW:
				if ((channel>=0)&&(channel<5))
				{
					if (pinModes[channel]!=PIN_IS_OUTPUT)
					{
						pinModes[channel]=PIN_IS_OUTPUT;
						if (streamDigital&((unsigned long)1<<channel))
							streamDigital^=(unsigned long)1<<channel;
					}
					set_digital_output(IO_D0+channel, LOW);
					writePacket();
				}
				break;
				//analog stream
			case  SVP_SET_ANALOG_STREAM:
				if (readValuePacket())
				{
					streamAnalog = value;
					writeValuePacket(value);

					for (channel=0;channel<13;channel++) lastAnalog[channel]=-1;
				}
				break;
				//battery stream
			case  SVP_SET_BATTERY_STREAM:
				if (readValuePacket())
				{
					batteryStream = value;
					writeValuePacket(value);
					lastBattery=-1;
				}
				break;
				//lcd
			case  SVP_SET_LCD:
				readLCDPacket();
				break;
			case  SVP_PLAY:
				readNotesPacket();
				break;
			}
		}

		for (channel=0;channel<13;channel++)
		{
			if (streamAnalog&(1<<channel))
			{
				if (channel<=7)
					value = analog_read(channel);
				else
					if (channel==8)
						value = analog_read(CHANNEL_A);
					else
						if (channel==9)
							value = analog_read(CHANNEL_B);
						else
							if (channel==10)
								value = analog_read(CHANNEL_C);
							else
								if (channel==11)
									value = analog_read(CHANNEL_D);
								else
									if (channel==12)
										value = analog_read(TRIMPOT);

				// only send value if it has changed
				if (value!=lastAnalog[channel])
				{
					command = SVP_ANALOG_STREAM;
					writeValuePacket(value);
					lastAnalog[channel]=value;
				}
			}
		}

		if (streamDigital)
		{
			unsigned long digitalValue=0;
			for (channel=0;channel<5;channel++)
			{
				if (streamDigital&((unsigned long)1<<channel))
				{
					if (pinModes[channel]!=PIN_IS_INPUT)
					{
						set_digital_input(IO_D0+channel, PULL_UP_ENABLED);
						pinModes[channel]=PIN_IS_INPUT;
					}

					// special case for green led
					if (channel==4)
					{
						if (is_digital_input_high(IO_D7))
							digitalValue |= 1<<channel;
					}
					else
					{
						if (is_digital_input_high(IO_D0+channel))
							digitalValue |= 1<<channel;
					}
				}
			}

			// only send value if it has changed
			if (digitalValue!=lastDigital)
			{
				command = SVP_DIGITAL_STREAM;
				channel = 0;
				writeValuePacket(digitalValue);
				lastDigital=digitalValue;
			}
		}

		unsigned char button = get_single_debounced_button_press(ANY_BUTTON);

		if (button & TOP_BUTTON)  // if top button pressed
		{
			if (!lastButtonState[0])
			{
				command = SVP_SET_BUTTON;
				channel = 0;
				writeValuePacket(1);
				lastButtonState[0]=1;
			}
		}


		if (button & MIDDLE_BUTTON)  // if middle button pressed
		{
			if (!lastButtonState[1])
			{
				command = SVP_SET_BUTTON;
				channel = 1;
				writeValuePacket(1);
				lastButtonState[1]=1;
			}
		}


		if (button & BOTTOM_BUTTON)  // if bottom button pressed
		{
			if (!lastButtonState[2])
			{
				command = SVP_SET_BUTTON;
				channel = 2;
				writeValuePacket(1);
				lastButtonState[2]=1;
			}
		}

		button = get_single_debounced_button_release(ANY_BUTTON);

		if (button & TOP_BUTTON)  // if top button released
		{
			if (lastButtonState[0])
			{
				command = SVP_SET_BUTTON;
				channel = 0;
				writeValuePacket(0);
				lastButtonState[0]=0;
			}
		}

		if (button & MIDDLE_BUTTON)  // if middle button released
		{
			if (lastButtonState[1])
			{
				command = SVP_SET_BUTTON;
				channel = 1;
				writeValuePacket(0);
				lastButtonState[1]=0;
			}
		}

		if (button & BOTTOM_BUTTON)  // if bottom button released
		{
			if (lastButtonState[2])
			{
				command = SVP_SET_BUTTON;
				channel = 2;
				writeValuePacket(0);
				lastButtonState[2]=0;
			}
		}

		if (batteryStream)
		{
			int battery = read_battery_millivolts();
			if ((int)(battery/100)!=(int)(lastBattery/100))
			{
				command = SVP_BATTERY_STREAM;
				channel = 0;
				writeValuePacket(battery);
				lastBattery = battery;
			}
		}
	}
}
